/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: Compiler.cc,v 1.36 2002/06/05 08:18:31 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef __MINGW32__
#include <assert.h>
#endif

#include "Compiler.hh"

#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/Firewall.hh"
#include "fwbuilder/Network.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/CustomService.hh"
#include "fwbuilder/Policy.hh"
#include "fwbuilder/Rule.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/InterfacePolicy.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/XMLTools.hh"
#include "fwbuilder/FWException.hh"
#include "fwbuilder/Group.hh"

#include <iostream.h>
#include <iomanip>
#include <algorithm>
#include <functional>
 
using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

FWCompilerException::FWCompilerException(Rule *r,const string &err) : FWException(err)
{
    rule=r;
}

Firewall* fwcompiler::findFirewallByName(FWObjectDatabase* db,
                                         const string &name) throw(FWException)
{
    FWObject *fwgrp=db->getById( FWObjectDatabase::db->std.FirewallsId , true );
    assert(fwgrp!=NULL);

    FWObject::const_iterator i=
	std::find_if( fwgrp->begin(), fwgrp->end(), FWObjectNameEQPredicate(name) );

    if (i==fwgrp->end()) throw FWException("Firewall object '"+name+"' not found");

    return (i==fwgrp->end())?NULL:Firewall::cast( *i );
}



Compiler::~Compiler() {}

int Compiler::cache_objects(FWObject *o)
{
    if ( !o->getId().empty() )  cacheObj(o);

    int n=0;
    for (FWObject::iterator i=o->begin(); i!=o->end(); ++i) {
        n=n+1+cache_objects((*i));
    }
    return n;
}

void Compiler::cacheObj(libfwbuilder::FWObject *o)
{
    objcache[o->getId()]=o;
}


int Compiler::prolog() 
{
/* caching firewall interfaces */
    FWObjectTypedChildIterator j=fw->findByType(Interface::TYPENAME);
    for ( ; j!=j.end(); ++j ) {
        Interface *interface_=Interface::cast(*j);
        fw_interfaces[interface_->getId()]=interface_;
    }
    fw_id=fw->getId();

    fwopt = fw->getOptionsObject();

/* caching all objects */

    cache_objects( dbcopy );

    return 0;
}

void Compiler::epilog()
{
    cerr << "Compiler::epilog must be overloaded\n";
    exit(1);
}

void Compiler::_init(FWObjectDatabase *_db, const string &fwobjectname)
{ 
    initialized=false;
    _cntr_=1; 

    fw=NULL; 
    temp_ruleset=NULL; 
    combined_ruleset=NULL;

    debug=0;
    debug_rule=-1;
    verbose=true;
    
    dbcopy=new FWObjectDatabase(*_db);  // copies entire tree

//    *(FWObject*)(&dbcopy) = *_db;

    FWObject *fwgrp=dbcopy->getById( FWObjectDatabase::db->std.FirewallsId , true );
    assert(fwgrp!=NULL);

    FWObject *o;
    list<FWObject*> l1=fwgrp->getByType(Firewall::TYPENAME);
    for (list<FWObject*>::iterator i=l1.begin(); i!=l1.end(); ++i) {
	o=(*i);
	if (o && o->getName()==fwobjectname) {
	    fw=Firewall::cast(o);
	    break;
	}
    }

    if (fw==NULL) {
	cerr << "Firewall object '" << fwobjectname << "' not found \n";
	exit(1);
    }

}

Compiler::Compiler(FWObjectDatabase *_db, 
		   const string &fwobjectname)
{
    _init(_db,fwobjectname);
}

Compiler::Compiler(FWObjectDatabase *_db, 
		   const string &fwobjectname,
		   OSConfigurator *_oscnf)
{
    _init(_db,fwobjectname);
    osconfigurator=_oscnf;
}

string Compiler::createRuleLabel(const string &interface_,int rule_num)
{
    return createRuleLabel("",interface_,rule_num);
}

string Compiler::createRuleLabel(const string &txt,const string &interface_,int rule_num)
{
    ostringstream  str;
    
    str << txt;

    str << rule_num;
    if (!interface_.empty()) str << "(" << interface_ << ")";

    return str.str();
}

void Compiler::abort(const string &errstr)
{
    error(errstr);
    exit(1);
}

void Compiler::error(const string &errstr)
{
    cout << flush;
    cerr << "Error (" << myPlatformName() << "): ";
    cerr << errstr << endl;
}

void Compiler::warning(const string &warnstr)
{
    cout << flush;
    cerr << "Warning (" << myPlatformName() << "): ";
    cerr << warnstr << endl;
}




string Compiler::getUniqueRuleLabel()
{
    char str[64];
    sprintf(str,"R_%d",_cntr_);
    _cntr_++;
    return str;
}



void Compiler::compile()
{
    assert(fw);
    assert(combined_ruleset);

}



void Compiler::_expand_group_recursive(FWObject *o,list<FWObject*> &ol)
{
    if (Group::cast( o )!=NULL) {
	for (FWObject::iterator i2=o->begin(); i2!=o->end(); ++i2) {
	    FWObject *o1= *i2;
	    if (FWReference::cast(o1)!=NULL) o1=getCachedObj(o1->getStr("ref"));
	    assert(o1);

	    _expand_group_recursive(o1,ol);
	}
    } else {
	o->ref();
	ol.push_back( o );
    }
}

/**
 *   object 's' here is really src or dst or srv. Its children objects
 *   should all be references
 */
void Compiler::expandGroupsInRuleElement(RuleElement *s)
{
    list<FWObject*> cl;
    for (FWObject::iterator i1=s->begin(); i1!=s->end(); ++i1) {
	FWObject *o= *i1;
	if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
	assert(o);

	_expand_group_recursive(o,cl);
    }

    s->clearChildren();
    s->setAnyElement();

    for(FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) {
	s->addRef( *i1 );
    }
}

/**
 * internal: scans children of 's' and, if found host or firewall with
 * multiple interfaces, replaces reference to that host or firewall
 * with a set of references to its interfaces. Argument 's' should be 
 * a pointer at either src or dst in the rule
 *
 */
void Compiler::_expandAddr(Rule *rule,FWObject *s) 
{
    Interface *rule_iface = fw_interfaces[rule->getStr("interface_id")];
    bool on_loopback= ( rule_iface && rule_iface->isLoopback() );

    list<FWObject*> cl;
    for (FWObject::iterator i1=s->begin(); i1!=s->end(); ++i1) {

	FWObject *o= *i1;
	if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
	assert(o);

	FWObjectTypedChildIterator j=o->findByType(Interface::TYPENAME);

	if (j==j.end()) cl.push_back( o );
	else {

/*
 * Special case is loopback interface - skip it, but only if this rule is
 * not attached to loopback!
 */
	    for ( ; j!=j.end(); ++j ) {
		Interface *interface_=Interface::cast(*j);
//		if ( interface_->isDyn() ) continue;
		if ( ! on_loopback && interface_->isLoopback() ) continue;
		cl.push_back(interface_);
	    }
	}
    }

    if ( ! cl.empty() )     {
	s->clearChildren();

	for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  {
	    s->addRef( *i1 );
	}
    }
}



void Compiler::_expandAddressRanges(Rule *rule,FWObject *s) 
{
    list<FWObject*> cl;
    for (FWObject::iterator i1=s->begin(); i1!=s->end(); ++i1) {

	FWObject *o= *i1;
	if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
	assert(o!=NULL);

	if (AddressRange::cast(o)==NULL) cl.push_back(o);
	else {
/* 
 * TODO : use smarter algorithm - replace address range with
 * combination of hosts and network objects to reduce number of
 * temporary objects we create
 */
	    IPAddress a1=AddressRange::cast(o)->getRangeStart();
	    IPAddress a2=AddressRange::cast(o)->getRangeEnd();
	    ObjectGroup *hosts=ObjectGroup::cast( dbcopy->getById(dbcopy->std.HostsId,true) );

	    guint32 distance=ntohl( a2.to32BitInt()-a1.to32BitInt() );
	    if (distance>32) {
		ostringstream  str;
		str << "Using AddressRange object in the rule "
		    << rule->getStr("rule_label")
		    << " is inefficient (address range will be replaced with "
		    << distance
		    << " individual addresses). Consider using network objects instead.";
		warning(str.str());
	    }

	    while (a1<=a2) {
		Host *h= Host::cast(dbcopy->create(Host::TYPENAME,true) );
		h->setStr("address", a1.toString() );
		h->setName(string("%")+a1.toString()+string("%") );
		cacheObj(h); // to keep cache consistent
		hosts->add(h);
		cl.push_back(h);

		a1=a1+1;
	    }
	}

    }
    if ( ! cl.empty() )     {
	s->clearChildren();

	for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  {
	    s->addRef( *i1 );
	}
    }
}

void Compiler::normalizePortRange(int &rs,int &re)
{
    if (rs<0) rs==0;
    if (re<0) re==0;
    if (rs!=0 && re==0) re=rs;
}


/**
 * returns true if :
 * 1. obj1 is the same as obj2, or 
 * 2. any child of obj2 or 
 * 3. its address matches that of any obj2's interfaces, or 
 * 4. its address matches broadcast address of one of the interfaces
 * 5. address of obj1 is a broadcast (255.255.255.255)
 */
bool Compiler::complexMatch(Address *obj1,Address *obj2)
{
    if (obj1->getId()==obj2->getId() ||
        obj2->getById( obj1->getId() , true )!=NULL ) return true;

    IPAddress obj1_addr=obj1->getAddress();
    if (obj1_addr.isBroadcast()) return true;

    FWObjectTypedChildIterator j=obj2->findByType(Interface::TYPENAME);
    for ( ; j!=j.end(); ++j ) {
        Interface *iface=Interface::cast(*j);
        if ( iface->getAddress()==obj1_addr ) return true;
        if ( ! iface->isDyn() ) {
            IPNetwork n( iface->getAddress() , iface->getNetmask() );
            if (n.getBroadcastAddress()==obj1_addr)  return true;
        }
    }

    return false;
}

/**
 * This method finds interface of obj2 (which is usually
 * firewall object, but not necessarily so) which is connected
 * to the subnet on which obj1 is located.
 */
Interface* Compiler::findInterfaceFor(Address *obj1,Address *obj2)
{
    FWObjectTypedChildIterator j=obj2->findByType(Interface::TYPENAME);
    for ( ; j!=j.end(); ++j ) {
	Interface *iface=Interface::cast(*j);
	assert(iface);
        if ( iface->isDyn() ) {
            if (iface->getId()      == obj1->getId() )      return iface;
        } else {

            if (iface->getAddress() == obj1->getAddress() ) return iface;

            if (Network::cast(obj1)!=NULL) {
                IPNetwork n1( obj1->getAddress() , Network::cast(obj1)->getNetmask() );
                if (n1.belongs( iface->getAddress() ) ) return iface; 
            }

/* this is the network interface is sitting on */
            IPNetwork n2( iface->getAddress() , iface->getNetmask() );
            if ( n2.belongs( obj1->getAddress() ) )      return iface;

        }
    }
    return NULL;
}

void Compiler::debugRule()
{
    for(FWObject::iterator i=combined_ruleset->begin(); i!=combined_ruleset->end(); i++) {
	Rule *rule = Rule::cast( *i );

        if ( rule->getPosition()==debug_rule ) {
            cout << debugPrintRule(rule);
            cout << endl;
        }
    }
}

/*
 *  basic rule printing, not very useful. This method is overloaded in
 *  derived classes
 */
string Compiler::debugPrintRule(libfwbuilder::Rule *rule)
{
    return rule->getStr("rule_label");
}



/**
 *  adds rule processor to the chain and, if debugging is ON, also
 *  adds rule processor "Debug" after that. Do not add Debug after
 *  certain processors, such as SimplePrintProgress
 */
void Compiler::add(BasicRuleProcessor* rp) 
{ 
    rule_processors.push_back(rp); 
    if (debug_rule>=0  && dynamic_cast<simplePrintProgress*>(rp)==NULL) 
        rule_processors.push_back(new Debug());
}

void Compiler::runRuleProcessors()
{
    list<BasicRuleProcessor*>::iterator i=rule_processors.begin();
    (*i)->setContext(this);
    list<BasicRuleProcessor*>::iterator j=i;
    ++i;
    for ( ; i!=rule_processors.end(); ++i) {
        (*i)->setContext(this);
        (*i)->setDataSource( (*j) );
        j=i;
    }

    while( (*j)->processNext() );
}

Compiler::Begin::Begin(const std::string &n) : BasicRuleProcessor(n) 
{
    init=false;
};

bool Compiler::Begin::processNext()
{
    assert(compiler!=NULL);
    if (!init) {
        for (FWObject::iterator i=compiler->combined_ruleset->begin();
             i!=compiler->combined_ruleset->end(); ++i) {
            Rule *rule=Rule::cast(*i);
            tmp_queue.push_back( rule );
        }
        init=true;
        return true;
    }
    return false;
}

bool Compiler::printTotalNumberOfRules::processNext()
{
    assert(compiler!=NULL);
    assert(prev_processor!=NULL);

    slurp();
    if (tmp_queue.size()==0) return false;
    if (compiler->verbose) cout << " processing " << tmp_queue.size() << " rules" << endl << flush;
    return true;
}

bool Compiler::Debug::processNext()
{
    assert(compiler!=NULL);
    assert(prev_processor!=NULL);

    slurp();
    if (tmp_queue.size()==0) return false;

    if (compiler->debug_rule>=0) {
        string n=prev_processor->getName();
        cout << endl << flush;
        cout << "--- "  << n << " " << setw(74-n.length()) << setfill('-') << "-" << flush;

        for (std::deque<Rule*>::iterator i=tmp_queue.begin(); i!=tmp_queue.end(); ++i) {
            Rule *rule=Rule::cast(*i);
            if ( rule->getPosition()==compiler->debug_rule ) {
                cout << compiler->debugPrintRule(rule) << flush;
                cout << endl << flush;
            }
        }
    }
    return true;
}

bool Compiler::simplePrintProgress::processNext()
{
    Rule *rule=prev_processor->getNextRule(); if (rule==NULL) return false;

    string rl=rule->getStr("rule_label");
    if (rl!=current_rule_label) {
            
        if (compiler->verbose) {
            string s=" rule "+rl+"\n";
            cout << s << flush;
        }

        current_rule_label=rl;
    }

    tmp_queue.push_back(rule);
    return true;
}


